# -*- coding: utf-8 -*-

import time

import requests

from mozci.data.base import DataSource
from mozci.errors import ContractNotFilled
from mozci.util.req import get_session

BUGBUG_BASE_URL = "https://bugbug.herokuapp.com"
DEFAULT_API_KEY = "mozci"
DEFAULT_RETRY_TIMEOUT = 9 * 60  # seconds
DEFAULT_RETRY_INTERVAL = 10  # seconds


class BugbugTimeoutException(Exception):
    pass


def get_schedules(branch, rev):
    session = get_session()
    session.headers.update({"X-API-KEY": DEFAULT_API_KEY})

    timeout = DEFAULT_RETRY_TIMEOUT
    interval = DEFAULT_RETRY_INTERVAL
    # On try there is no fallback and pulling is slower, so we allow bugbug more
    # time to compute the results.
    # See https://github.com/mozilla/bugbug/issues/1673.
    if branch == "try":
        timeout += int(timeout / 3)

    attempts = int(round(timeout / interval))
    for _ in range(0, attempts):
        response = session.get(f"{BUGBUG_BASE_URL}/push/{branch}/{rev}/schedules")

        if response.ok and response.status_code != 202:
            return response.json()

        time.sleep(interval)

    # Reaching this code means that either the last attempt ended with an error or timed out
    response.raise_for_status()
    raise BugbugTimeoutException(
        "Timed out waiting for result from Bugbug HTTP Service"
    )


class BugbugSource(DataSource):
    """Queries bugbug.herokuapp.com for schedules data about pushes."""

    name = "bugbug"
    supported_contracts = ("push_test_selection_data",)

    def run_push_test_selection_data(self, branch, rev):
        try:
            results = get_schedules(branch, rev)
        except requests.exceptions.HTTPError:
            raise ContractNotFilled(
                self.name,
                "push_test_selection_data",
                "could not retrieve schedules from Bugbug",
            )

        return results



# -*- coding: utf-8 -*-

from datetime import datetime

import requests
from loguru import logger
from taskcluster import Index
from taskcluster.exceptions import TaskclusterRestFailure

from mozci.data.base import DataSource
from mozci.errors import ContractNotFilled
from mozci.util import taskcluster


class TaskclusterSource(DataSource):
    """Queries Taskcluster for data about tasks."""

    name = "taskcluster"
    supported_contracts = (
        "push_tasks",
        "push_test_selection_data",
        "push_existing_classification",
    )

    def to_ms(self, datestring, fmt="%Y-%m-%dT%H:%M:%S.%fZ"):
        dt = datetime.strptime(datestring, fmt)
        return int(dt.timestamp() * 1000)

    def run_push_tasks(self, branch, rev):
        try:
            decision_task_id = taskcluster.find_task_id(
                f"gecko.v2.{branch}.revision.{rev}.taskgraph.decision"
            )
        except requests.exceptions.HTTPError as e:
            # If the decision task was not indexed, it means it was broken. So we can
            # assume we didn't run any task for this push.
            if e.response.status_code == 404:
                logger.warning(f"Decision task broken in {rev} on {branch}")
                return []

            raise

        task_data = taskcluster.get_task(decision_task_id)

        results = taskcluster.get_tasks_in_group(task_data["taskGroupId"])

        tasks = []
        for result in results:
            # Skip the decision task.
            if result["status"]["taskId"] == decision_task_id:
                continue

            # Skip "Action" tasks.
            if result["task"]["metadata"]["name"].startswith("Action"):
                continue

            task = {
                "id": result["status"]["taskId"],
                "label": result["task"]["metadata"]["name"],
                "tags": result["task"]["tags"],
                "state": result["status"]["state"],
            }
            tier = result["task"]["extra"].get("treeherder", {}).get("tier")
            if tier:
                task["tier"] = tier

            # Use the latest run (earlier ones likely had exceptions that
            # caused an automatic retry).
            if task["state"] != "unscheduled":
                run = result["status"]["runs"][-1]

                # Normalize the state to match treeherder's values.
                if task["state"] == "failed":
                    task["state"] = "completed"

                # Derive a result from the reasonResolved.
                reason = run.get("reasonResolved")
                if reason == "completed":
                    task["result"] = "passed"
                elif reason in ("canceled", "superseded", "failed"):
                    task["result"] = reason
                elif reason:
                    task["result"] = "exception"
                else:
                    # Task is not finished, so there is no result yet.
                    assert task["state"] in ("pending", "running", "exception")

                # Compute duration.
                if "started" in run and "resolved" in run:
                    task["duration"] = self.to_ms(run["resolved"]) - self.to_ms(
                        run["started"]
                    )

            tasks.append(task)

        return tasks

    def run_push_test_selection_data(self, branch, rev):
        task_name = f"gecko.v2.{branch}.revision.{rev}.taskgraph.decision"
        try:
            decision_task_id = taskcluster.find_task_id(task_name)
        except requests.exceptions.HTTPError as e:
            # If the decision task was not indexed, it means it was broken. So we can
            # assume we didn't run any task for this push.
            if e.response.status_code == 404:
                logger.warning(f"Decision task broken in {rev} on {branch}")

            raise ContractNotFilled(
                self.name,
                "push_test_selection_data",
                f"could not retrieve decision task '{task_name}'",
            )

        try:
            results = taskcluster.get_artifact(
                decision_task_id, "public/bugbug-push-schedules.json"
            )
        except requests.exceptions.HTTPError:
            raise ContractNotFilled(
                self.name,
                "push_test_selection_data",
                "could not retrieve schedules from cache",
            )

        return results

    def run_push_existing_classification(self, branch, rev, environment):
        # Non-production environments are exposed in sub routes
        route_prefix = (
            "project.mozci.classification"
            if environment == "production"
            else f"project.mozci.{environment}.classification"
        )

        # We use buildUrl and manual requests.get instead of directly findArtifactFromTask from the taskcluster library
        # because the taskcluster library fails with redirects (https://github.com/taskcluster/taskcluster/issues/4998).
        try:
            # Proxy authentication does not seem to work here
            index = Index({"rootUrl": taskcluster.COMMUNITY_TASKCLUSTER_ROOT_URL})
            url = index.buildUrl(
                "findArtifactFromTask",
                f"{route_prefix}.{branch}.revision.{rev}",
                "public/classification.json",
            )
        except TaskclusterRestFailure as e:
            raise ContractNotFilled(
                self.name,
                "push_existing_classification",
                f"Failed to load existing classification for {branch} {rev}: {e}",
            )

        try:
            r = requests.get(url, allow_redirects=True)
            r.raise_for_status()
        except requests.exceptions.HTTPError as e:
            raise ContractNotFilled(
                self.name,
                "push_existing_classification",
                f"Failed to load existing classification for {branch} {rev}: {e}",
            )

        try:
            return r.json()["push"]["classification"]
        except KeyError:
            raise ContractNotFilled(
                self.name, "push_existing_classification", "Invalid classification data"
            )



# -*- coding: utf-8 -*-
import os
import re
from collections import namedtuple
from itertools import groupby
from typing import Any, Dict

import requests
from cleo import Command
from loguru import logger

from mozci import config
from mozci.push import make_push_objects
from mozci.util.defs import TASK_FINAL_STATES
from mozci.util.taskcluster import (
    COMMUNITY_TASKCLUSTER_ROOT_URL,
    find_task_id,
    index_current_task,
    list_dependent_tasks,
    list_indexed_tasks,
    notify_matrix,
)

BackfillTask = namedtuple("BackfillTask", ["task_id", "label", "th_symbol", "state"])

NOTIFICATION_BACKFILL_GROUP_COMPLETED = "Backfill tasks associated to the Treeherder symbol {th_symbol} for push [{push.branch}/{push.rev}](https://treeherder.mozilla.org/jobs?group_state=expanded&repo={push.branch}{tochange}{fromchange}{searchstr}) are all in a final state."


class CheckBackfillsCommand(Command):
    """
    Check if backfills on last pushes are finished and notify Sheriffs when they are.

    check-backfills
        {--branch=autoland : Branch the pushes belongs to (e.g autoland, try, etc).}
        {--nb-pushes=40 : Number of recent pushes to retrieve for the check.}
        {--environment=testing : Environment in which the analysis is running (testing, production, ...)}

    The command will execute the following workflow:
      1. Retrieve the last <--nb-pushes> Pushes on branch <--branch>
      => Then for each Push:
      2. Check if it has any associated actions triggered by Treeherder, using Taskcluster indexation
      3. Find potential backfill tasks
      4. Group them by backfill groups
      => For each backfill group on the Push:
      5. Check if all backfill tasks in this group are completed
      6. If so (and if the notification wasn't already sent), send a notification on Matrix alerting that this backfill group is completed
      7. Add the current task in a dedicated index to avoid sending multiple times the same notification
    """

    def handle(self) -> None:
        branch = self.option("branch")
        environment = self.option("environment")
        matrix_room = config.get("matrix-room-id")
        current_task_id = os.environ.get("TASK_ID")

        try:
            nb_pushes = int(self.option("nb-pushes"))
        except ValueError:
            self.line("<error>Provided --nb-pushes should be an int.</error>")
            exit(1)

        self.line("<comment>Loading pushes...</comment>")
        self.pushes = make_push_objects(nb=nb_pushes, branch=branch)
        nb_pushes = len(self.pushes)

        to_notify: Dict[str, Dict[str, Any]] = {}
        for index, push in enumerate(self.pushes, start=1):
            self.line(
                f"<comment>Processing push {index}/{nb_pushes}: {push.push_uuid}</comment>"
            )
            backfill_tasks = []

            try:
                indexed_tasks = list_indexed_tasks(
                    f"gecko.v2.{push.branch}.revision.{push.rev}.taskgraph.actions"
                )
            except requests.exceptions.HTTPError as e:
                self.line(
                    f"<error>Couldn't fetch indexed tasks on push {push.push_uuid}: {e}</error>"
                )
                continue

            for indexed_task in indexed_tasks:
                task_id = indexed_task["taskId"]
                try:
                    children_tasks = list_dependent_tasks(task_id)
                except requests.exceptions.HTTPError as e:
                    self.line(
                        f"<error>Couldn't fetch dependent tasks of indexed task {task_id} on push {push.push_uuid}: {e}</error>"
                    )
                    continue

                for child_task in children_tasks:
                    task_section = child_task.get("task", {})
                    task_action = task_section.get("tags", {}).get("action", "")
                    # We are looking for the Treeherder symbol because Sheriffs are
                    # only interested in backfill-tasks holding the '-bk' suffix in TH
                    th_symbol = (
                        task_section.get("extra", {})
                        .get("treeherder", {})
                        .get("symbol", "")
                    )
                    status = child_task.get("status", {})
                    if task_action == "backfill-task" and th_symbol.endswith("-bk"):
                        assert status.get(
                            "taskId"
                        ), "Missing taskId attribute in backfill task status"
                        label = task_section.get("tags", {}).get(
                            "label"
                        ) or task_section.get("metadata", {}).get("name")
                        assert (
                            label
                        ), "Missing label attribute in backfill task tags or name attribute in backfill task metadata"
                        assert status.get(
                            "state"
                        ), "Missing state attribute in backfill task status"
                        backfill_tasks.append(
                            BackfillTask(
                                status["taskId"], label, th_symbol, status["state"]
                            )
                        )
                    else:
                        logger.debug(
                            f"Skipping non-backfill task {status.get('taskId')}"
                        )

            def group_key(task):
                return task.th_symbol

            # Sorting backfill tasks by their Treeherder symbol
            backfill_tasks = sorted(backfill_tasks, key=group_key)
            # Grouping ordered backfill tasks by their associated Treeherder symbol
            for th_symbol, tasks_iter in groupby(backfill_tasks, group_key):
                if th_symbol not in to_notify:
                    to_notify[th_symbol] = {
                        "newest_push": None,
                        "backfill_tasks": set(),
                    }

                # make_push_objects returns the latest pushes in chronological order from oldest to newest
                # We only need to store the newest Push that appeared for this Treeherder symbol
                to_notify[th_symbol]["newest_push"] = push
                # Storing all backfill tasks for this symbol across multiple pushes
                to_notify[th_symbol]["backfill_tasks"].update(tasks_iter)

        for th_symbol, data in to_notify.items():
            logger.debug("Preparing notification for the Treeherder symbol {th_symbol}")

            all_backfill_tasks = data["backfill_tasks"]
            # Checking that all backfill tasks for this symbol are in a "final" state
            if not all(task.state in TASK_FINAL_STATES for task in all_backfill_tasks):
                logger.debug(
                    f"Not all backfill tasks for the Treeherder symbol {th_symbol} are in a final state, not notifying now."
                )
                continue

            newest_push = data["newest_push"]
            index_path = f"project.mozci.check-backfill.{environment}.{newest_push.branch}.{newest_push.rev}.{th_symbol}"
            try:
                find_task_id(index_path, root_url=COMMUNITY_TASKCLUSTER_ROOT_URL)
            except requests.exceptions.HTTPError:
                pass
            else:
                logger.debug(
                    f"A notification was already sent for the backfill tasks associated to the Treeherder symbol {th_symbol}."
                )
                continue

            try:
                parents = [
                    parent for parent in newest_push._iterate_parents(max_depth=20)
                ]
            except Exception as e:
                logger.debug(
                    f"Failed to load the last twenty parent pushes for push {newest_push.push_uuid}, because: {e}."
                )
                parents = None

            logger.debug("Generating notification text")

            cleaned_label = re.sub(
                r"(-e10s|-1proc)?(-\d+)?$", "", all_backfill_tasks.pop().label
            )
            notification = NOTIFICATION_BACKFILL_GROUP_COMPLETED.format(
                th_symbol=th_symbol,
                push=newest_push,
                tochange=f"&tochange={newest_push.child.rev}",
                fromchange=f"&fromchange={parents[-1].rev}" if parents else "",
                searchstr=f"&searchStr={cleaned_label}",
            )

            if not matrix_room:
                self.line(
                    f"<comment>A notification should be sent for the backfill tasks associated to the Treeherder symbol {th_symbol} but no matrix room was provided in the secret.</comment>"
                )
                logger.debug(f"The notification: {notification}")
                continue

            logger.debug("Sending Matrix notification")

            # Sending a notification to the Matrix channel defined in secret
            notify_matrix(
                room=matrix_room,
                body=notification,
            )

            if not current_task_id:
                self.line(
                    f"<comment>The current task should be indexed in {index_path} but TASK_ID environment variable isn't set.</comment>"
                )
                continue

            logger.debug("Indexing current task")

            # Populating the index with the current task to prevent sending the notification once again
            index_current_task(
                index_path,
                root_url=COMMUNITY_TASKCLUSTER_ROOT_URL,
            )



# -*- coding: utf-8 -*-
import csv
import datetime
import itertools
import json
import os
import time
import uuid
from multiprocessing import Pool

from cleo import Command
from loguru import logger

from mozci.console.commands.push import (
    parse_and_log_details,
    prepare_for_analysis,
    retrieve_sheriff_intermittents,
    retrieve_sheriff_reals,
)
from mozci.push import Push, PushStatus, make_push_objects
from mozci.util.defs import INTERMITTENT_CLASSES

BASE_OUTPUT_DIR = (
    os.path.dirname(os.path.realpath(__file__)) + "/classify_batch_results"
)
DAYS_FROM_TODAY = 30
PARAMETERS_NAMES = [
    "intermittent_confidence_threshold",
    "real_confidence_threshold",
    "use_possible_regressions",
    "unknown_from_regressions",
    "consider_children_pushes_configs",
    "cross_config_counts",
    "consistent_failures_counts",
]
PARAMETERS_COMBINATIONS = [
    dict(zip(PARAMETERS_NAMES, parameters_values))
    for parameters_values in itertools.product(
        [0.5, 0.7, 0.9],  # intermittent_confidence_threshold
        [0.7, 0.8, 0.9],  # real_confidence_threshold
        [True, False],  # use_possible_regressions
        [True, False],  # unknown_from_regressions
        [True, False],  # consider_children_pushes_configs
        [(2, 2), (2, 3), (2, 5)],  # cross_config_counts
        [(2, 3), (2, 5)],  # consistent_failures_counts
    )
]
MAXIMUM_PROCESSES = os.cpu_count() or 1


def retrieve_push_and_prepare_for_analysis(push):
    push = Push(push["rev"], branch=push["branch"])
    prepare_for_analysis(push)


def _serialize_regressions(regressions):
    return {
        group: [{"task_id": task.id, "label": task.label} for task in failing_tasks]
        for group, failing_tasks in regressions.items()
    }


def create_json_file(push, run_id, classification_name, regressions):
    to_save = {
        "push": {
            "id": push.push_uuid,
            "classification": classification_name,
        },
        "failures": {
            "real": _serialize_regressions(regressions.real),
            "intermittent": _serialize_regressions(regressions.intermittent),
            "unknown": _serialize_regressions(regressions.unknown),
        },
    }

    with open(f"{BASE_OUTPUT_DIR}/{push.id}/{run_id}.json", "w") as file:
        json.dump(to_save, file, indent=2)


def run_combinations_for_push(push):
    push = Push(push["rev"], branch=push["branch"])

    push_dir = f"{BASE_OUTPUT_DIR}/{push.id}"
    if not os.path.exists(push_dir):
        os.makedirs(push_dir)

    csv_rows = []
    for parameters in PARAMETERS_COMBINATIONS:
        run_id = uuid.uuid4()

        start = time.time()
        try:
            classification, regressions = push.classify(**parameters)
            end = time.time()

            # Only save results to a JSON file if the execution was successful
            classification_name = classification.name
            create_json_file(push, run_id, classification_name, regressions)
        except Exception as e:
            end = time.time()
            classification_name = "SYSTEM_ERROR"
            logger.error(
                f"An error occurred during the classification of push {push.push_uuid}: {e}"
            )

        csv_rows.append(
            {
                "run_uuid": run_id,
                "push_uuid": push.push_uuid,
                **parameters,
                "classification": classification_name,
                "time_spent": round(end - start, 3),
                "now": datetime.datetime.now(),
            }
        )

    return csv_rows


class BatchClassificationCommand(Command):
    """
    Run the classification algorithm with various parameters combinations for all submitted pushes within the last 30 days

    classify
        {--workers= : Number of workers to use in order to parallelize the executions.}
    """

    csv_header = (
        ["run_uuid", "push_uuid"]
        + PARAMETERS_NAMES
        + ["classification", "time_spent", "now"]
    )

    def handle(self) -> None:
        # Default value from a constant if nothing is provided
        workers_count = MAXIMUM_PROCESSES

        workers_option = self.option("workers")
        if workers_option:
            try:
                workers_count = int(workers_option)
            except ValueError:
                self.line("<error>Provided --workers should be an int.</error>")
                exit(1)

            if workers_count > MAXIMUM_PROCESSES:
                self.line(
                    f"<comment>Parallelization over {workers_count} workers was requested but only {MAXIMUM_PROCESSES} CPUs are available, falling back to using only {MAXIMUM_PROCESSES} workers.</comment>"
                )
                workers_count = MAXIMUM_PROCESSES

        pushes = self.retrieve_pushes()
        self.line(
            f"<info>{len(pushes)} pushes will be classified using {len(PARAMETERS_COMBINATIONS)} parameters combinations.</info>"
        )

        with Pool(workers_count) as pool:
            # Populate the cache + Clean up potential biases
            pool.map(retrieve_push_and_prepare_for_analysis, pushes)

        if not os.path.exists(BASE_OUTPUT_DIR):
            os.makedirs(BASE_OUTPUT_DIR)

        with open(
            BASE_OUTPUT_DIR + "/all_executions.csv", "w", encoding="UTF8", newline=""
        ) as f:
            writer = csv.DictWriter(f, fieldnames=self.csv_header)
            writer.writeheader()

            with Pool(workers_count) as pool:
                # Each time an execution ends (execution = all combinations for a single push),
                # its result will be appended to the CSV
                for csv_rows in pool.imap(run_combinations_for_push, pushes):
                    writer.writerows(csv_rows)

    def retrieve_pushes(self):
        now = datetime.datetime.today()
        to_date = now.strftime("%Y-%m-%d")
        from_date = (now - datetime.timedelta(days=DAYS_FROM_TODAY)).strftime(
            "%Y-%m-%d"
        )
        pushes = make_push_objects(
            from_date=from_date, to_date=to_date, branch="autoland"
        )
        return [{"rev": push.rev, "branch": push.branch} for push in pushes]


class BatchEvaluationCommand(Command):
    """
    Evaluate and aggregate the results produced by the classify_batch_execution script

    evaluate
    """

    csv_header = [
        "configuration",
        "total_pushes",
        "classify_errors",
        "evaluate_errors",
        "correct_good",
        "wrong_good",
        "correct_bad",
        "wrong_bad",
        "total_real",
        "correct_real",
        "wrong_real",
        "missed_real",
        "total_intermittent",
        "correct_intermittent",
        "wrong_intermittent",
        "missed_intermittent",
    ]

    def handle(self) -> None:
        results_path = BASE_OUTPUT_DIR + "/all_executions.csv"
        assert os.path.exists(results_path) and os.path.isfile(
            results_path
        ), "The CSV containing results from all classify executions doesn't exist"

        with open(results_path, newline="") as f:
            reader = csv.DictReader(f)
            configs = self.parse_csv(reader)

        # Write evaluation results per config in a new CSV file
        with open(
            BASE_OUTPUT_DIR + "/evaluation_of_all_executions.csv",
            "w",
            encoding="UTF8",
            newline="",
        ) as f:
            writer = csv.DictWriter(f, fieldnames=self.csv_header)
            writer.writeheader()

            csv_rows = []
            for config, results in configs.items():
                results.update({"configuration": config})
                csv_rows.append(results)

            writer.writerows(csv_rows)

    def evaluate_push_failures(
        self, push_id, run_uuid, push_group_summaries, sheriff_data
    ):
        """
        Evaluate real/intermittent failures on a Push
        """
        with open(f"{BASE_OUTPUT_DIR}/{push_id}/{run_uuid}.json") as f:
            classify_results = json.load(f)
            predicted_reals = classify_results["failures"]["real"].keys()
            predicted_intermittents = classify_results["failures"][
                "intermittent"
            ].keys()

        evaluation = {}
        evaluation.update(
            # Evaluate real failures that were predicted
            parse_and_log_details(
                push_group_summaries,
                sheriff_data["reals"],
                {"fixed by commit"},
                predicted_groups=predicted_reals,
                ignore_pending_conflicting=True,
                suffix="_real",
            )[0]
        )
        evaluation.update(
            # Evaluate intermittent failures that were predicted
            parse_and_log_details(
                push_group_summaries,
                sheriff_data["intermittents"],
                set(INTERMITTENT_CLASSES),
                predicted_groups=predicted_intermittents,
                ignore_pending_conflicting=True,
                suffix="_intermittent",
            )[0]
        )

        return evaluation

    def parse_csv(self, reader):
        """
        Parse the CSV file and evaluate each classify result for a specific configuration and Push
        """
        pushes_group_summaries = {}
        pushes_sheriff_data = {}

        configs = {}
        for i, row in enumerate(reader, start=2):
            self.line(f"Evaluating line {i} of the CSV")
            config = ",".join(
                ["=".join([param, row[param]]) for param in PARAMETERS_NAMES]
            )

            if config not in configs:
                # Initialize all counts at 0
                configs[config] = {key: 0 for key in self.csv_header[1:]}

            configs[config]["total_pushes"] += 1

            if row["classification"] == "SYSTEM_ERROR":
                # No need to evaluate since an error happened during Push classification
                configs[config]["classify_errors"] += 1
                continue

            push_uuid = row["push_uuid"]
            try:
                branch, rev = push_uuid.split("/")
                push = Push(rev, branch=branch)

                if row["classification"] == PushStatus.GOOD.name:
                    if push.backedout:
                        configs[config]["wrong_good"] += 1
                    else:
                        configs[config]["correct_good"] += 1
                elif row["classification"] == PushStatus.BAD.name:
                    if push.backedout:
                        configs[config]["correct_bad"] += 1
                    else:
                        configs[config]["wrong_bad"] += 1

                if push_uuid not in pushes_group_summaries:
                    pushes_group_summaries[push_uuid] = push.group_summaries

                if push_uuid not in pushes_sheriff_data:
                    # By storing sheriffs annotations here, we will only have to run
                    # retrieve_sheriff_data once per push, since it can take a long
                    # time to fetch everything, this helps speed up the evaluation
                    pushes_group_summaries, sheriff_reals = retrieve_sheriff_reals(
                        pushes_group_summaries, push
                    )
                    (
                        pushes_group_summaries,
                        sheriff_intermittents,
                    ) = retrieve_sheriff_intermittents(pushes_group_summaries, push)
                    pushes_sheriff_data[push_uuid] = {
                        "reals": sheriff_reals,
                        "intermittents": sheriff_intermittents,
                    }

                evaluation = self.evaluate_push_failures(
                    push.id,
                    row["run_uuid"],
                    pushes_group_summaries[push_uuid],
                    pushes_sheriff_data[push_uuid],
                )
                for key, value in evaluation.items():
                    configs[config][key] += value

            except Exception as e:
                configs[config]["evaluate_errors"] += 1
                self.line(
                    f"<error>An error occurred during the evaluation of the classify execution with run_uuid {row['run_uuid']}: {e}</error>"
                )

        return configs


class BatchExecutionCommands(Command):
    """
    Contains commands that operate on lots of pushes using lots of classification algorithm configurations.

    batch-execution
    """

    commands = [
        BatchClassificationCommand(),
        BatchEvaluationCommand(),
    ]

    def handle(self):
        return self.call("help", self._config.name)



# -*- coding: utf-8 -*-

FAILURE_CLASSES = ("not classified", "fixed by commit")
INTERMITTENT_CLASSES = ("intermittent", "autoclassified intermittent")
TASK_FINAL_STATES = ("completed", "failed", "exception")



# -*- coding: utf-8 -*-


# push exceptions


class BasePushException(Exception):
    def __init__(self, rev, branch, msg):
        self.rev = rev
        self.branch = branch
        self.msg = f"Error with push '{rev}' on {branch}: {msg}"


class PushNotFound(BasePushException):
    """Raised when the requested push does not exist."""

    def __init__(self, reason, *args, **kwargs):
        kwargs["msg"] = "push not found '{reason}'!"
        super(PushNotFound, self).__init__(*args, **kwargs)


class ChildPushNotFound(BasePushException):
    """Raised when a suitable child push could not be found."""

    def __init__(self, reason, *args, **kwargs):
        kwargs["msg"] = f"child push not found '{reason}'!"
        super(ChildPushNotFound, self).__init__(*args, **kwargs)


class ParentPushNotFound(BasePushException):
    """Raised when a suitable parent push could not be found."""

    def __init__(self, reason, *args, **kwargs):
        kwargs["msg"] = f"parent push not found '{reason}'!"
        super(ParentPushNotFound, self).__init__(*args, **kwargs)


# task exceptions


class BaseTaskException(Exception):
    def __init__(self, id, label, msg):
        self.id = id
        self.label = label
        self.msg = f"Error with task '{id}' ({label}): {msg}"


class ArtifactNotFound(BaseTaskException):
    """Raised when the requested task artifact does not exist."""

    def __init__(self, artifact, *args, **kwargs):
        kwargs["msg"] = f"artifact '{artifact}' does not exist!"
        self.artifact = artifact
        super(ArtifactNotFound, self).__init__(*args, **kwargs)


class TaskNotFound(BaseTaskException):
    """Raised when a Task id or index could not be found."""

    def __init__(self, *args, **kwargs):
        kwargs["msg"] = "task not found!"
        super(TaskNotFound, self).__init__(*args, **kwargs)


# data exceptions


class BaseDataException(Exception):
    pass


class ContractNotFilled(BaseDataException):
    """Raised when a source was unable to fulfill a contract."""

    def __init__(self, source, contract, reason):
        self.source = source
        self.contract = contract
        self.reason = reason
        self.msg = f"Unable to fulfill '{contract}' with '{source}' source: {reason}"


class ContractNotFound(BaseDataException):
    """Raised when a requested contract does not exist."""

    def __init__(self, contract):
        self.contract = contract
        self.msg = f"Contract '{contract}' is not defined!"


class SourcesNotFound(BaseDataException):
    """Raised when no sources were able to fulfill a contract."""

    def __init__(self, contract, context):
        self.contract = contract
        self.context = context
        self.msg = f"No registered sources were able to fulfill '{contract}'!"


class InvalidSource(BaseDataException):
    """Raised when a source is malformed."""

    def __init__(self, source, reason):
        self.source = source
        self.reason = reason
        self.msg = f"'{source}' source is malformed: {reason}"


class MissingDataError(ValueError):
    """Raised when a source returns no data."""

    pass



# -*- coding: utf-8 -*-
import os
import pickle
import shutil
import tarfile
import tempfile
import threading
from distutils.dir_util import copy_tree

import taskcluster
from cachy.contracts.store import Store
from cachy.serializers import Serializer
from cachy.stores import FileStore, NullStore  # noqa
from loguru import logger

from mozci.util.req import get_session
from mozci.util.taskcluster import get_taskcluster_options


def extract_tar_zst(path, dest):
    import zstandard

    dctx = zstandard.ZstdDecompressor()
    with open(path, "rb") as f:
        with dctx.stream_reader(f) as reader:
            with tarfile.open(mode="r|", fileobj=reader) as tar:
                tar.extractall(dest)


class SeededFileStore(FileStore):
    RESEED_KEY = "mozci:SeededFileStore:reseed"

    def __init__(self, config):
        """A FileStore instance that allows pre-seeding the cache from an archive
        downloaded from a URL.

        Example configuration:

            file = {
                driver = "seeded-file",
                path = "/path/to/cache",
                url = "https://example.com/cache.tar.gz"
            }

        Configuration can include:

            path: Path on the local file system to store the cache.
            url: Where to download the preseed data from.
            archive_relpath: Path within the archive pointing to the root of the cache data.
            reseed_interval: Time in minutes after which the data will be
                              seeded again (defaults to no reseeding).

        Supported archive formats include `.zip`, `.tar`, `.tar.gz`, `.tar.bz2` or `.tar.zst`.
        """
        self._url = config["url"]
        self._reseed_interval = config.get("reseed_interval", 0)
        self._archive_relpath = config.get("archive_relpath")

        self._session = get_session()

        kwargs = {
            "directory": config["path"],
        }
        if "hash_type" in config:
            kwargs["hash_type"] = config["hash_type"]

        super(SeededFileStore, self).__init__(**kwargs)

    def seed(self):
        """Download and extract the seed data to the cache directory."""
        logger.info(
            f"Seeding mozci cache at {self._directory} with contents from {self._url}"
        )
        self._create_cache_directory(self._directory)

        filename = self._url.split("/")[-1]
        with self._session.get(self._url, stream=True) as r:
            r.raise_for_status()

            with tempfile.TemporaryDirectory() as tempdir:
                with tempfile.NamedTemporaryFile(suffix=filename) as fh:
                    shutil.copyfileobj(r.raw, fh)

                    if filename.endswith(".tar.zst"):
                        extract_tar_zst(fh.name, tempdir)
                    else:
                        shutil.unpack_archive(fh.name, tempdir)

                path = tempdir
                if self._archive_relpath:
                    path = os.path.join(tempdir, self._archive_relpath)

                copy_tree(path, self._directory)

        # Make sure we don't reseed again for another 'reseed_interval' minutes.
        self.put(self.RESEED_KEY, False, self._reseed_interval)

    def get(self, key):
        reseed = super(SeededFileStore, self).get(self.RESEED_KEY)
        if reseed is None:
            self.seed()

        return super(SeededFileStore, self).get(key)


class RenewingFileStore(FileStore):
    def __init__(self, config, retention):
        """A FileStore instance that renews items in the cache when they are
        accessed again.
        """
        kwargs = {
            "directory": config["path"],
        }
        if "hash_type" in config:
            kwargs["hash_type"] = config["hash_type"]

        self.retention = retention

        super(RenewingFileStore, self).__init__(**kwargs)

    def get(self, key):
        value = super(RenewingFileStore, self).get(key)
        if value is None:
            return None

        self.put(key, value, self.retention)
        return value


def get_s3_credentials(bucket, prefix):
    auth = taskcluster.Auth(get_taskcluster_options())
    response = auth.awsS3Credentials("read-write", bucket, prefix)
    return response["credentials"]


S3_CLIENTS_LOCK = threading.Lock()
S3_CLIENTS = {}


def get_s3_client(bucket, prefix):
    import boto3

    with S3_CLIENTS_LOCK:
        if (bucket, prefix) not in S3_CLIENTS:
            credentials = get_s3_credentials(bucket, prefix)
            S3_CLIENTS[(bucket, prefix)] = boto3.client(
                "s3",
                aws_access_key_id=credentials["accessKeyId"],
                aws_secret_access_key=credentials["secretAccessKey"],
                aws_session_token=credentials["sessionToken"],
            )

        return S3_CLIENTS[(bucket, prefix)]


def destroy_s3_client(bucket, prefix):
    with S3_CLIENTS_LOCK:
        del S3_CLIENTS[(bucket, prefix)]


class S3Store(Store):
    def __init__(self, config):
        """A Store instance that stores items in S3."""
        self._bucket = config["bucket"]
        self._prefix = config["prefix"]

    @property
    def _client(self):
        return get_s3_client(self._bucket, self._prefix)

    def _key(self, key):
        return os.path.join(self._prefix, key)

    def _retry_if_expired(self, op):
        import botocore

        try:
            return op()
        except botocore.exceptions.ClientError:
            destroy_s3_client(self._bucket, self._prefix)
            return op()

    def _forget(self, key):
        self._client.delete_object(Bucket=self._bucket, Key=self._key(key))
        # Should return False if the item was not present in the cache, but we
        # don't care.
        return True

    def forget(self, key):
        return self._retry_if_expired(lambda: self._forget(key))

    def _get(self, key):
        import botocore

        # Copy the object onto itself to extend its expiration.
        try:
            head = self._client.head_object(Bucket=self._bucket, Key=self._key(key))

            # Change its metadata, or Amazon will complain.
            metadata = head["Metadata"]
            metadata["id"] = "1" if "id" in metadata and metadata["id"] == "0" else "0"

            self._client.copy_object(
                Bucket=self._bucket,
                CopySource={"Bucket": self._bucket, "Key": self._key(key)},
                Key=self._key(key),
                Metadata=metadata,
                MetadataDirective="REPLACE",
            )
        except botocore.exceptions.ClientError as ex:
            if ex.response["Error"]["Code"] == "404":
                return None
            raise

        response = self._client.get_object(Bucket=self._bucket, Key=self._key(key))
        data = response["Body"].read()
        try:
            return self.unserialize(data)
        except Exception:
            # The object is broken, let's delete it.
            self.forget(key)
            return None

    def get(self, key):
        return self._retry_if_expired(lambda: self._get(key))

    def _put(self, key, value):
        self._client.put_object(
            Body=self.serialize(value), Bucket=self._bucket, Key=self._key(key)
        )

    def put(self, key, value, minutes):
        self._retry_if_expired(lambda: self._put(key, value))


class CompressedPickleSerializer(Serializer):
    def __init__(self):
        import zstandard

        self.compressor = zstandard.ZstdCompressor(
            level=zstandard.MAX_COMPRESSION_LEVEL
        )
        self.decompressor = zstandard.ZstdDecompressor()

    def serialize(self, data):
        return self.compressor.compress(pickle.dumps(data))

    def unserialize(self, data):
        return pickle.loads(self.decompressor.decompress(data))



# -*- coding: utf-8 -*-
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import datetime
import os
from urllib.parse import urlencode

import markdown2
import taskcluster
import taskcluster_urls as liburls
from loguru import logger

from mozci.util import yaml
from mozci.util.req import get_session

PRODUCTION_TASKCLUSTER_ROOT_URL = "https://firefox-ci-tc.services.mozilla.com"
queue = taskcluster.Queue({"rootUrl": PRODUCTION_TASKCLUSTER_ROOT_URL})
COMMUNITY_TASKCLUSTER_ROOT_URL = "https://community-tc.services.mozilla.com"


def _do_request(url, force_get=False, use_put=False, **kwargs):
    session = get_session()
    if kwargs and not force_get:
        if not use_put:
            response = session.post(url, **kwargs)
        else:
            response = session.put(url, **kwargs)
    else:
        response = session.get(url, stream=True, **kwargs)
    if response.status_code >= 400:
        # Consume content before raise_for_status, so that the connection can be
        # reused.
        response.content
    response.raise_for_status()
    return response


def _handle_artifact(path, response):
    if path.endswith(".json"):
        return response.json()
    if path.endswith(".yml"):
        return yaml.load_stream(response.text)
    return response


def get_artifact_url(task_id, path, root_url=PRODUCTION_TASKCLUSTER_ROOT_URL):
    return liburls.api(
        root_url,
        "queue",
        "v1",
        f"task/{task_id}/artifacts/{path}",
    )


def get_artifact(task_id, path, root_url=PRODUCTION_TASKCLUSTER_ROOT_URL):
    """
    Returns the artifact with the given path for the given task id.

    If the path ends with ".json" or ".yml", the content is deserialized as,
    respectively, json or yaml, and the corresponding python data (usually
    dict) is returned.
    For other types of content, a file-like object is returned.
    """
    response = _do_request(get_artifact_url(task_id, path, root_url=root_url))

    return _handle_artifact(path, response)


def list_artifacts(task_id):
    return queue.listLatestArtifacts(task_id)["artifacts"]


def get_index_url(index_path, root_url=PRODUCTION_TASKCLUSTER_ROOT_URL):
    return liburls.api(
        root_url,
        "index",
        "v1",
        f"task/{index_path}",
    )


def find_task_id(index_path, use_proxy=False, root_url=PRODUCTION_TASKCLUSTER_ROOT_URL):
    response = _do_request(get_index_url(index_path, root_url=root_url))
    return response.json()["taskId"]


def index_current_task(
    index_path,
    rank=0,
    expires=None,
    data={},
    root_url=PRODUCTION_TASKCLUSTER_ROOT_URL,
):
    if expires is None:
        expires = datetime.datetime.now() + datetime.timedelta(days=1 * 365)

    index_service = taskcluster.Index(get_taskcluster_options())
    index_service.insertTask(
        index_path,
        {
            "data": data,
            "expires": expires,
            "rank": rank,
            "taskId": os.environ["TASK_ID"],
        },
    )


def get_indexed_tasks_url(namespace, root_url=PRODUCTION_TASKCLUSTER_ROOT_URL):
    return liburls.api(
        root_url,
        "index",
        "v1",
        f"tasks/{namespace}",
    )


def list_indexed_tasks(namespace, root_url=PRODUCTION_TASKCLUSTER_ROOT_URL):
    url = get_indexed_tasks_url(namespace, root_url=root_url)
    token = False
    # Support pagination using continuation token
    while token is not None:
        extra_params = "?" + urlencode({"continuationToken": token}) if token else ""
        results = _do_request(url + extra_params).json()
        yield from results.get("tasks", [])
        token = results.get("continuationToken")


def get_task_url(task_id):
    return liburls.api(
        PRODUCTION_TASKCLUSTER_ROOT_URL, "queue", "v1", f"task/{task_id}"
    )


def get_task(task_id, use_proxy=False):
    return queue.task(task_id)


def get_dependent_tasks_url(task_id, root_url=PRODUCTION_TASKCLUSTER_ROOT_URL):
    return liburls.api(root_url, "queue", "v1", f"task/{task_id}/dependents")


def list_dependent_tasks(task_id, root_url=PRODUCTION_TASKCLUSTER_ROOT_URL):
    url = get_dependent_tasks_url(task_id, root_url=root_url)
    token = False
    # Support pagination using continuation token
    while token is not None:
        extra_params = "?" + urlencode({"continuationToken": token}) if token else ""
        results = _do_request(url + extra_params).json()
        yield from results.get("tasks", [])
        token = results.get("continuationToken")


def create_task(task_id, task):
    options = taskcluster.optionsFromEnvironment()
    options["rootUrl"] = PRODUCTION_TASKCLUSTER_ROOT_URL
    queue = taskcluster.Queue(options)
    return queue.createTask(task_id, task)


def get_tasks_in_group(group_id):
    tasks = []

    def _save_tasks(response):
        tasks.extend(response["tasks"])

    queue.listTaskGroup(group_id, paginationHandler=_save_tasks)

    return tasks


def get_taskcluster_options():
    """
    Helper to get the Taskcluster setup options
    according to current environment (local or Taskcluster)
    """
    options = taskcluster.optionsFromEnvironment()
    proxy_url = os.environ.get("TASKCLUSTER_PROXY_URL")

    if proxy_url is not None:
        # Always use proxy url when available
        options["rootUrl"] = proxy_url

    if "rootUrl" not in options:
        # Always have a value in root url
        options["rootUrl"] = "https://community-tc.services.mozilla.com"

    return options


def notify_email(subject, content, emails):
    """
    Send an email to all provided email addresses
    using Taskcluster notify service
    """
    if not emails:
        logger.warning("No email address available in configuration")
        return

    notify_service = taskcluster.Notify(get_taskcluster_options())
    for idx, email in enumerate(emails):
        try:
            notify_service.email(
                {
                    "address": email,
                    "subject": f"Mozci | {subject}",
                    "content": content,
                }
            )
        except Exception as e:
            logger.error(
                f"Failed to send the report by email to address n°{idx} ({email}): {e}"
            )
            raise


def notify_matrix(body, room):
    """
    Send a message on the provided Matrix room
    using Taskcluster notify service
    """
    if not room:
        logger.warning("No Matrix room available in configuration")
        return

    formatted_body = markdown2.markdown(body)

    notify_service = taskcluster.Notify(get_taskcluster_options())
    try:
        notify_service.matrix(
            {
                "roomId": room,
                "body": body,
                "format": "org.matrix.custom.html",
                "formattedBody": formatted_body,
                "msgtype": "m.text",
            }
        )
    except Exception as e:
        logger.error(f"Failed to send the report on the Matrix room {room}: {e}")
        raise



# -*- coding: utf-8 -*-
import copy
import os
from collections.abc import Mapping
from pathlib import Path

import taskcluster
from appdirs import user_config_dir
from cachy import CacheManager
from loguru import logger
from taskcluster.helper import load_secrets
from tomlkit import parse

from mozci.util.cache_stores import (
    CompressedPickleSerializer,
    NullStore,
    RenewingFileStore,
    S3Store,
    SeededFileStore,
)
from mozci.util.taskcluster import get_taskcluster_options


def merge_to(source, dest):
    """
    Merge dict and arrays (override scalar values).

    Keys from source override keys from dest, and elements from lists in source
    are appended to lists in dest.

    Args:
        source (dict): to copy from
        dest (dict): to copy to (modified in place)
    """
    for key, value in source.items():

        if key not in dest:
            dest[key] = value
            continue

        # Merge dict
        if isinstance(value, dict) and isinstance(dest[key], dict):
            merge_to(value, dest[key])
            continue

        if isinstance(value, list) and isinstance(dest[key], list):
            dest[key] = dest[key] + value
            continue

        dest[key] = value

    return dest


def flatten(d, prefix=""):
    if prefix:
        prefix += "."

    result = []
    for key, value in d.items():
        if isinstance(value, dict):
            result.extend(flatten(value, prefix=f"{prefix}{key}"))
        elif isinstance(value, (set, list)):
            vstr = "\n".join([f"    {i}" for i in value])
            result.append(f"{prefix}{key}=\n{vstr}")
        else:
            result.append(f"{prefix}{key}={value}")

    return sorted(result)


class CustomCacheManager(CacheManager):
    def __init__(self, cache_config):
        super_config = {
            k: v
            for k, v in cache_config.items()
            # We can't pass the serializer config to the CacheManager constructor,
            # as it tries to resolve it but we have not had a chance to register it
            # yet.
            if k != "serializer"
        }
        super_config.setdefault("stores", {"null": {"driver": "null"}})
        super(CustomCacheManager, self).__init__(super_config)

        for store, conf in self._config["stores"].items():
            if store != "null":
                logger.debug(f"Active cache store {store} with conf {conf}")

        self.extend("null", lambda driver: NullStore())
        self.extend("seeded-file", SeededFileStore)
        self.extend(
            "renewing-file",
            lambda config: RenewingFileStore(config, cache_config["retention"]),
        )
        self.extend("s3", S3Store)

        self.register_serializer("compressedpickle", CompressedPickleSerializer())

        # Now we can manually set the serializer we wanted.
        self._serializer = self._resolve_serializer(
            cache_config.get("serializer", "pickle")
        )


class Configuration(Mapping):
    DEFAULT_CONFIG_PATH = Path(user_config_dir("mozci")) / "config.toml"
    TASKCLUSTER_CONFIG_SECRET = os.environ.get(
        "TASKCLUSTER_CONFIG_SECRET"
    ) or os.environ.get("TASKCLUSTER_SECRET")
    DEFAULTS = {
        "merge": {
            "retriggerable-backfillable-task-names": [],
            "autoclassification": {
                "enabled": False,
                "test-suite-names": [],
                "failure-types": [],
            },
            "cache": {"retention": 1440},
        },  # minutes
        "replace": {
            "data_sources": [
                "hgmo",
                "taskcluster",
                "treeherder_client",
                "errorsummary",
                "bugbug",
            ],
            "verbose": 0,
            "tier": 2,
        },
    }

    locked = False

    def __init__(self, path=None):
        self.path = Path(
            path or os.environ.get("MOZCI_CONFIG_PATH") or self.DEFAULT_CONFIG_PATH
        )

        self._config = copy.deepcopy(self.DEFAULTS["merge"])
        if self.TASKCLUSTER_CONFIG_SECRET is not None:
            # Load configuration from Taskcluster
            self.merge(self.load_from_secret())
        elif self.path.is_file():
            # Load configuration from local file
            with open(self.path, "r") as fh:
                content = fh.read()
                self.merge(parse(content)["mozci"])
        else:
            logger.warning(f"Configuration path {self.path} is not a file.")

        for k, v in self.DEFAULTS["replace"].items():
            self._config.setdefault(k, v)

        self.cache = CustomCacheManager(self._config["cache"])
        self.locked = True

        # Check auto classification settings
        assert isinstance(self._config["autoclassification"]["enabled"], bool)
        assert isinstance(self._config["autoclassification"]["test-suite-names"], list)

        assert isinstance(self._config["retriggerable-backfillable-task-names"], list)

    def __len__(self):
        return len(self._config)

    def __iter__(self):
        return iter(self._config)

    def __getitem__(self, key):
        return self._config[key]

    def __getattr__(self, key):
        if key in vars(self):
            return vars(self)[key]
        return self.__getitem__(key)

    def __setattr__(self, key, value):
        if self.locked:
            raise AttributeError(
                "Don't set attributes directly, use `config.set(key=value)` instead."
            )
        super(Configuration, self).__setattr__(key, value)

    def set(self, **kwargs):
        """Set data on the config object."""
        self._config.update(kwargs)

    def merge(self, other):
        """Merge data into config (updates dicts and lists instead of
        overwriting them).

        Args:
            other (dict): Dictionary to merge configuration with.
        """
        merge_to(other, self._config)

    def update(self, config):
        """
        Update the configuration object with new parameters
        :param config: dict of configuration
        """
        for k, v in config.items():
            if v is not None:
                self._config[k] = v

        object.__setattr__(self, "cache", CustomCacheManager(self._config["cache"]))

    def dump(self):
        return "\n".join(flatten(self._config))

    def load_from_secret(self):
        """
        Load configuration from a Taskcluster secret
        """
        assert self.TASKCLUSTER_CONFIG_SECRET, "Missing TASKCLUSTER_CONFIG_SECRET"
        logger.info(
            f"Loading configuration from secret {self.TASKCLUSTER_CONFIG_SECRET}"
        )

        return load_secrets(
            secrets_service=taskcluster.Secrets(get_taskcluster_options()),
            secret_name=self.TASKCLUSTER_CONFIG_SECRET,
        )


config = Configuration()



# -*- coding: utf-8 -*-

from dataclasses import dataclass
from textwrap import dedent
from typing import Tuple, Union

import validx as v

from mozci.task import FailureType, TestTask


@dataclass
class Contract:
    name: str
    description: str
    validate_in: v.Dict
    validate_out: Union[v.Dict, v.List]


_contracts: Tuple[Contract, ...] = (
    Contract(
        name="push_tasks",
        description="Data about the tasks that ran on a given push.",
        validate_in=v.Dict(
            {
                "branch": v.Str(),
                "rev": v.Str(),
            }
        ),
        validate_out=v.List(
            v.Dict(
                {
                    "id": v.Str(),
                    "label": v.Str(),
                    "state": v.Str(
                        options=[
                            "completed",
                            "running",
                            "pending",
                            "unscheduled",
                            "exception",
                        ]
                    ),
                    "tags": v.Dict(extra=(v.Str(), v.Str())),
                    "tier": v.Int(),
                    "duration": v.Int(),
                    "result": v.Str(
                        options=[
                            "passed",
                            "failed",
                            "exception",
                            "canceled",
                            "superseded",
                        ]
                    ),
                },
                optional=["duration", "result", "tier"],
            )
        ),
    ),
    Contract(
        name="push_tasks_classifications",
        description=dedent(
            """
            Return classifications on the tasks that ran on a given push. Tasks
            without a classification need not be present.
        """
        ),
        validate_in=v.Dict(
            {
                "branch": v.Str(),
                "rev": v.Str(),
            }
        ),
        validate_out=v.Dict(
            extra=(
                v.Str(),
                v.Dict(
                    {
                        "classification": v.Str(
                            options=[
                                "autoclassified intermittent",
                                "infra",
                                "intermittent",
                                "expected fail",
                                "fixed by commit",
                                "not classified",
                            ],
                        ),
                        "classification_note": v.Str(),
                    },
                    optional=["classification_note"],
                ),
            )
        ),
    ),
    Contract(
        name="push_revisions",
        description="Data from the VCS about a given push.",
        validate_in=v.Dict(
            {
                "from_date": v.Str(),
                "to_date": v.Str(),
                "branch": v.Str(),
            }
        ),
        validate_out=v.List(
            v.Dict(
                {
                    "pushid": v.Int(),
                    "date": v.Int(),
                    "revs": v.List(
                        v.Dict(
                            {
                                "author": v.Str(),
                                "branch": v.Str(),
                                "desc": v.Str(),
                                "files": v.List(v.Str()),
                                "node": v.Str(),
                                "parents": v.List(v.Str()),
                                "tags": v.List(v.Str()),
                            }
                        )
                    ),
                }
            )
        ),
    ),
    Contract(
        name="test_task_groups",
        description="A dict of test groups and their results and durations for a given TestTask.",
        validate_in=v.Dict(
            {
                "branch": v.Str(),
                "rev": v.Str(),
                "task": v.Type(TestTask),
            }
        ),
        validate_out=v.Dict(
            # TODO: 'minlen=1' can be added to v.Str(), the group name, once we stop seeing groups with empty names.
            # TODO: 'nullable=True' can be removed once https://github.com/mozilla/mozci/issues/662 is fixed.
            extra=(v.Str(), v.Tuple(v.Bool(), v.Int(nullable=True)))
        ),
    ),
    Contract(
        name="test_task_errors",
        description="A list of errors for a given TestTask.",
        validate_in=v.Dict(
            {
                "task": v.Type(TestTask),
            }
        ),
        validate_out=v.List(v.Str(minlen=1)),
    ),
    Contract(
        name="test_task_failure_types",
        description="A list of failures with their associated type grouped by test group for a given TestTask.",
        validate_in=v.Dict(
            {
                "task_id": v.Str(),
            }
        ),
        validate_out=v.Dict(
            extra=(
                v.Str(minlen=1),
                v.List(v.Tuple(v.Str(minlen=1), v.Type(FailureType))),
            )
        ),
    ),
    Contract(
        name="push_test_selection_data",
        description="Test and build CI Tasks that should be scheduled for a given push",
        validate_in=v.Dict(
            {
                "branch": v.Str(),
                "rev": v.Str(),
            }
        ),
        validate_out=v.Dict(
            {
                "config_groups": v.Dict(
                    extra=(v.Str(), v.List(v.Str(minlen=1))),
                ),
                "groups": v.Dict(
                    extra=(v.Str(), v.Float()),
                ),
                "known_tasks": v.List(v.Str(minlen=1)),
                "reduced_tasks": v.Dict(
                    extra=(v.Str(), v.Float()),
                ),
                "reduced_tasks_higher": v.Dict(
                    extra=(v.Str(), v.Float()),
                ),
                "tasks": v.Dict(
                    extra=(v.Str(), v.Float()),
                ),
            }
        ),
    ),
    Contract(
        name="pushes",
        description="List available pushes",
        validate_in=v.Dict(
            {
                "branch": v.Str(),
                "nb": v.Int(),
            }
        ),
        validate_out=v.List(
            v.Dict(
                {
                    "pushid": v.Int(),
                    "date": v.Int(),
                    "revs": v.List(v.Str()),
                }
            )
        ),
    ),
    Contract(
        name="push_existing_classification",
        description="Retrieve the pre-existing classification status of a given push",
        validate_in=v.Dict(
            {
                "branch": v.Str(),
                "rev": v.Str(),
                "environment": v.Str(),
            }
        ),
        validate_out=v.Str(options=["GOOD", "BAD", "UNKNOWN"]),
    ),
)


all_contracts = {c.name: c for c in _contracts}
